;; Generated package description from listen.el  -*- no-byte-compile: t -*-
(define-package "listen" "0.6pre0.20240308.134149" "Audio/Music player" '((emacs "29.1") (persist "0.6") (taxy "0.10") (taxy-magit-section "0.13")) :commit "1816e077cb4f0f9dec6dbb5bb17352ae835935bd" :authors '(("Adam Porter" . "adam@alphapapa.net")) :maintainer '("Adam Porter" . "adam@alphapapa.net") :keywords '("multimedia") :url "https://github.com/alphapapa/listen.el")
